/*
 * file: genps.c
 *
 * (c) Peter Kleiweg
 *     Mon Jun 29 09:37:38 1998
 */

#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define BUFSIZE 1024

#define POPSIZE 16
#define CODELEN 400

typedef enum {
    END, NUM, X, Y, SWAP, NEGX, NEGY, EXCH, DUP, ADD, DIV, MOD, MUL, SUB,
    ABS, NEG, SQRT, ATAN, COS, SIN, EXP, LN, NOISE, ERROR
} OPT;

typedef struct {
    int
        opcode;
    float
        op;
} CELL;

typedef struct {
    OPT
        opt;
    char const
        *name;
    int
        requires,      /* how many args should be on the stack, negative is special */
        changes;       /* how many args will be added to (removed from) stack */
} OPCODE;

OPCODE
    opcodes [] = {
            { END,  NULL,   1,  0 },   /* this one should be first */
            { NUM,  "NUM",  0,  1 },
            { X,    "X",    0,  1 },
            { Y,    "Y",    0,  1 },
            { SWAP, NULL,   0,  0 },
            { NEGX, NULL,   0,  0 },
            { NEGY, NULL,   0,  0 },
            { EXCH, "exch", 2,  0 },
            { DUP,  "dup",  1,  1 },
            { ADD,  "add",  2, -1 },
            { DIV,  "DIV",  2, -1 },
            { MOD,  "MOD",  2, -1 },
            { MUL,  "mul",  2, -1 },
            { SUB,  "sub",  2, -1 },
            { ABS,  "abs",  1,  0 },
            { NEG,  "neg",  1,  0 },
            { SQRT, "SQRT", 1,  0 },
            { ATAN, "ATAN", 2, -1 },
            { COS,  "COS",  1,  0 },
            { SIN,  "SIN",  1,  0 },
            { EXP,  "EXP",  2, -1 },
            { LN,   "LN",   1,  0 },
            { NOISE, "NOISE", 1, 0 }
        };

#define N_OPS 23

CELL
    pop [POPSIZE][CODELEN + 1],
    orgnew [CODELEN + 1];

char
    *programname,
    buffer [BUFSIZE + 1];

int
    noise = 1;

FILE
    *fpps,
    *fpps2;

void
    writecell (int i, FILE *fp),
    printcell (int i, FILE *fp),
    get_programname (char const *argv0),
    read_from_file (void),
    syntax (void),
    errit (char const *format, ...);

char
    header [] = "
/rectfill where {
    pop
} {
    /rectfill {
        gsave
            newpath
            4 2 roll moveto
            1 index 0 rlineto
            0 exch rlineto
            neg 0 rlineto
            closepath
            fill
        grestore
    } bind def
} ifelse
/X 0 def
/Y 0 def
/pixcol { } def
/PI 3.14159265358979323846 def
/E  2.7182818284590452354 def
/DIV {
    dup abs .00001 lt {
        0 lt {
            -.00001
        } {
            .00001
        } ifelse
    } if
    div
} bind def
/MOD {
    DIV
    dup floor sub
} bind def
/SQRT {
    dup 0 lt exch
    abs sqrt
    exch { neg } if
} bind def
/SIN {
    PI div 180 mul sin
} bind def
/COS {
    PI div 180 mul cos
} bind def
/ATAN {
    1 index 0 eq
    1 index 0 eq
    and {
        pop
    } {
        atan 180 div PI mul
    } ifelse
} bind def
/EXP {
    1 index 0 gt {
        exp
    } {
        pop
    } ifelse
} bind def
/LN {
    dup 0 lt exch
    abs
    dup .0001 lt { pop .0001 } if
    ln
    exch { neg } if
} bind def
/NOISE {
    rand sin .01 mul 1 add mul
} bind def
/X1 0 def
/X2 0 def
/Y1 0 def
/Y2 0 def
/D  1 def
/setcmykcolor where {
    pop
    /SETCOLOR {
        counttomark 0 eq { 0 } if
        counttomark 1 eq { dup } if
        counttomark 2 eq { 2 copy } if
        counttomark 3 eq { dup } if
        4 {
            4 1 roll
            dup floor sub
        } repeat
        .4 mul setcmykcolor
    } bind def
} {
    /SETCOLOR {
        counttomark 0 eq { 0 } if
        counttomark 1 eq { dup } if
        counttomark 2 eq { dup } if
        3 {
            3 1 roll
            dup floor sub
        } repeat
        setrgbcolor
    } bind def
} ifelse
/draw {
    D sub /Y2 exch def
    D sub /X2 exch def
    /Y1 exch def
    /X1 exch def
    X1 D X2 {
        /XX exch def
        XX X1 sub
        X2 X1 sub div
        2 mul
        1 sub
        /X exch def
        Y1 D Y2 {
            /YY exch def
            YY Y1 sub
            Y2 Y1 sub div
            2 mul
            1 sub
            /Y exch def
            mark
            { pixcol } stopped pop
            SETCOLOR
            cleartomark
            XX YY D D rectfill
        } for
    } for
} bind def
";

int main (int argc, char *argv [])
{
    int
        i,
        j,
        n,
        p [2],
        s1,
        s2;
    FILE
        *fp;
    time_t
        tp;

    get_programname (argv [0]);

    for (i = 1; i < argc; i++) {
        if (! strcmp (argv [i], "-n"))
            noise = 0;
        else
            syntax ();
    }

    printf (
        "\n"
        "%s - Genetic Art with PostScript\n"
        "\n"
        "(c) Peter kleiweg 1998\n"
        "\n"
        "http://www.let.rug.nl/~kleiweg/genart/\n"
        "\n",
        programname
    );

    if (noise)
        printf (
            "\nWARNING: Using noise\n\n"
            "Start program with option -n to suppress noise\n\n"
        );

    srand ((unsigned int) time (NULL));

    for (i = 0; i < POPSIZE; i++) {
        for (j = 0; j < CODELEN; j++) {
            pop [i][j].opcode = rand () % N_OPS;
	    if (! pop [i][j].opcode)
	        pop [i][j].opcode = rand () % N_OPS;
            pop [i][j].op = ((float) rand ()) / ((float)RAND_MAX) * 4.0 - 2.0;
	}
        pop [i][CODELEN].opcode = 0;
    }
    orgnew [CODELEN].opcode = 0;

    fpps = popen ("gs -r72 -g740x740 -", "w");
    if (! fpps)
        errit ("Opening pipe for reading: %s", strerror (errno));
    fpps2 = popen ("gs -r72 -g500x500 -", "w");
    if (! fpps2)
        errit ("Opening pipe for reading: %s", strerror (errno));
    fputs (header, fpps);
    fputs ("/D 2 def\n/Courier-Bold findfont 12 scalefont setfont\n", fpps);
    fputs (header, fpps2);
    fputs ("/D 2 def\n-45 -170 translate\n", fpps2);
    for (i = 0; i < POPSIZE; i++) {
        writecell (i, fpps);
        printcell (i, fpps);
        fprintf (
            fpps,
            "0 setgray\n"
            "%i %i moveto (%i) dup stringwidth pop neg 2 div 0 rmoveto show\n",
            (i % 4) * 180 + 100,
            (i / 4) * -180 + 548,
            i + 1
        );
    }
    fputs ("flushpage\n", fpps);
    fflush (fpps);
    fflush (fpps2);

    for (;;) {
        printf (
            "\nNumber: replace, (L)arge, (R)esolution, (S)ave, read from (F)ile, (Q)uit\n: "
        );
        fflush (stdout);
        fgets (buffer, BUFSIZE, stdin);
        if (buffer [0] == 'q' || buffer [0] == 'Q')
            break;
        if (buffer [0] == 'l' || buffer [0] == 'L') {
            printf ("Large image nr: ");
            fflush (stdout);
            fgets (buffer, BUFSIZE, stdin);
            i = atoi (buffer);
            if (i > 0 && i <= POPSIZE) {
                writecell (i - 1, fpps2);
                fprintf (fpps2, "45 170 545 670 draw flushpage\n");
                fflush (fpps2);
            }
            continue;
	}
        if (buffer [0] == 's' || buffer [0] == 'S') {
  	    for (i = 0; i < 10000; i++) {
                sprintf (buffer, "%s.%04i.ps", programname, i);
                fp = fopen (buffer, "r");
                if (fp) {
                    fclose (fp);
                    continue;
                }
                break;
            }
            fp = fopen (buffer, "w");
            if (! fp)
                errit ("Creating file \"%s\": %s", buffer, strerror (errno));
            fputs (
                "%!PS-Adobe-3.0 EPSF-3.0\n"
                "%%BoundingBox: 45 170 545 670\n"
                "%%Creator: ",
                fp
            );
            fprintf (fp, "%s", programname);
            fputs (
                ", (c) P. Kleiweg 1998\n"
                "%%CreationDate: ",
                fp
            );
            time (&tp);
            fputs (asctime (localtime (&tp)), fp);
            fputs ("%%EndComments\n\n32 dict begin\n", fp);
            fputs (header, fp);
            printf ("Saving figure as \"%s\"\n", buffer);
            i = 0;
            while (i < 1 || i > POPSIZE) {
                printf ("Figure to save: ");
                fflush (stdout);
                fgets (buffer, BUFSIZE, stdin);
                i = atoi (buffer);
	    }
            writecell (i - 1, fp);
            fputs (
                "45 170 545 670 draw\nend\nshowpage\n%%EOF\n",
                fp
            );
            fclose (fp);
            continue;
	}
        if (buffer [0] == 'r' || buffer [0] == 'R') {
            fputs ("Resolution, (L)arge or (S)mall: ", stdout);
            fflush (stdout);
            fgets (buffer, BUFSIZE, stdin);
            switch (buffer [0]) {
                case 'l':
                case 'L':
                    i = 1;
                    break;
                case 's':
                case 'S':
                    i = 2;
                    break;
                default:
                    i = 0;
	    }
            if (! i)
                continue;
            printf ("Resolution %s image (1-20): ", (i == 1) ? "large" : "small");
            fgets (buffer, BUFSIZE, stdin);
            j = atoi (buffer);
            if (j > 0 && j <= 20)
                fprintf ((i == 1) ? fpps2 : fpps, "/D %i def\n", j);
            continue;
        }
        if (buffer [0] == 'f' || buffer [0] == 'F') {
            read_from_file ();
            continue;
	}
        i = atoi (buffer);
        if (i > 0 && i <= POPSIZE) {
            printf ("Replacing nr %i\n", i);
            i--;
            printf ("Parent 1: ");
            fflush (stdout);
            fgets (buffer, BUFSIZE, stdin);
            p [0] = atoi (buffer);
            if (p [0] < 1 || p [0] > POPSIZE)
                continue;
            p [0]--;
            printf ("Parent 2: ");
            fflush (stdout);
            fgets (buffer, BUFSIZE, stdin);
            p [1] = atoi (buffer);
            if (p [1] < 1 || p [1] > POPSIZE)
                continue;
            p [1]--;
            for (n = 0; pop [p [0]][n].opcode; n++)
                ;
            j = rand () % (n ? n : CODELEN);
            s1 = 0;
            s2 = (rand () % 5) - 2;
            for (n = 0; n < CODELEN; n++)
  	        orgnew [n] = pop [p [(n < j) ? 0 : 1]][(n + ((n < j) ? s1 : s2)) % CODELEN];
            for (n = 0; n < 5; n++) {
                j = rand () % CODELEN;
                orgnew [j].opcode = rand () % N_OPS;
                orgnew [j].op = ((float) rand ()) / ((float)RAND_MAX) * 4.0 - 2.0; 
	    }
            memcpy (pop [i], orgnew, CODELEN * sizeof (CELL));
            writecell (i, fpps);
            printcell (i, fpps);
            fputs ("flushpage\n", fpps);
            fflush (fpps);
        }
    }

    fputs ("quit\n", fpps);
    pclose (fpps);
    fputs ("quit\n", fpps2);
    pclose (fpps2);
    return 0;
}

void read_from_file ()
{
    int
        i,
        p,
        nr,
        negx,
        negy;
    OPT
        type;
    FILE
        *fp;

    p = negx = negy = 0;

    printf ("\nReplace nr: ");
    fflush (stdout);
    fgets (buffer, BUFSIZE, stdin);
    nr = atoi (buffer);
    if (nr < 0 || nr > POPSIZE)
        return;
    nr--;
    printf ("\nRead from file: ");
    fflush (stdout);
    fgets (buffer, BUFSIZE, stdin);
    for (i = strlen (buffer) - 1; i >= 0; i--)
        if (isspace ((unsigned char) buffer [i]))
            buffer [i] = '\0';
        else
            break;
    if (! buffer [0])
        return;
    fp = fopen (buffer, "r");
    if (! fp) {
        printf ("Opening file \"%s\": %s\n", buffer, strerror (errno));
        return;
    }
    for (;;) {
        if (fgets (buffer, BUFSIZE, fp) == NULL) {
            printf ("/pixcol not found\n");
            fclose (fp);
            return;
	}
        if (strcmp (buffer, "/pixcol {\n") == 0)
            break;
    }
    for (;;) {
        if (fgets (buffer, BUFSIZE, fp) == 0) {
            fclose (fp);
            break;
        }
        for (i = strlen (buffer) - 1; i >= 0; i--)
            if (isspace ((unsigned char) buffer [i]))
                buffer [i] = '\0';
            else
                break;
        if (! strcmp (buffer, "} bind def")) {
            fclose (fp);
            break;
	}
        if (p >= CODELEN) {
            printf ("Too long for current codesize\n");
            fclose (fp);
            break;
	}
        if (! strcmp (buffer, "X"))
            type = X;
        else if (! strcmp (buffer, "Y"))
            type = Y;
        else if (! strcmp (buffer, "X neg"))
            type = NEGX;
        else if (! strcmp (buffer, "Y neg"))
            type = NEGY;
        else if (! strcmp (buffer, "exch"))
            type = EXCH;
        else if (! strcmp (buffer, "dup"))
            type = DUP;
        else if (! strcmp (buffer, "add"))
            type = ADD;
        else if (! strcmp (buffer, "DIV"))
            type = DIV;
        else if (! strcmp (buffer, "MOD"))
            type = MOD;
        else if (! strcmp (buffer, "mul"))
            type = MUL;
        else if (! strcmp (buffer, "sub"))
            type = SUB;
        else if (! strcmp (buffer, "abs"))
            type = ABS;
        else if (! strcmp (buffer, "neg"))
            type = NEG;
        else if (! strcmp (buffer, "SQRT"))
            type = SQRT;
        else if (! strcmp (buffer, "ATAN"))
            type = ATAN;
        else if (! strcmp (buffer, "COS"))
            type = COS;
        else if (! strcmp (buffer, "SIN"))
            type = SIN;
        else if (! strcmp (buffer, "EXP"))
            type = EXP;
        else if (! strcmp (buffer, "LN"))
            type = LN;
        else if (! strcmp (buffer, "NOISE"))
            type = NOISE;
        else if (buffer [0] == '-' || buffer [0] == '.' || isdigit ((unsigned char) buffer [0]))
            type = NUM;
        else
            type = ERROR;
        switch (type) {
            case X:
	        if (negx) {
                    pop [nr][p++].opcode = NEGX;
                    negx = 0;
		}
                if (p < CODELEN)
                    pop [nr][p++].opcode = X;
                break;
            case Y:
	        if (negy) {
                    pop [nr][p++].opcode = NEGY;
                    negy = 0;
		}
                if (p < CODELEN)
                    pop [nr][p++].opcode = Y;
                break;
            case NEGX:
	        if (! negx) {
                    pop [nr][p++].opcode = NEGX;
                    negx = 1;
		}
                if (p < CODELEN)
                    pop [nr][p++].opcode = X;
                break;
            case NEGY:
	        if (! negy) {
                    pop [nr][p++].opcode = NEGY;
                    negy = 1;
		}
                if (p < CODELEN)
                    pop [nr][p++].opcode = Y;
                break;
            case NOISE:
                if (noise)
                    pop [nr][p++].opcode = NOISE;
                else
                    printf ("Noise not used\n");
                break;
            case NUM:
                pop [nr][p].opcode = NUM;
                pop [nr][p++].op = atof (buffer);
                break;
            case ERROR:
                printf ("Not recognized: %s\n", buffer);
                break;
            default:
                pop [nr][p++].opcode = type;
        }
    }
    if (p < CODELEN)
        pop [nr][p].opcode = END;

    writecell (nr, fpps);
    printcell (nr, fpps);
    fputs ("flushpage\n", fpps);
    fflush (fpps);
 }

void writecell (int i, FILE *fp)
{
    int
        j,
        o,
        n,
        s,
        x,
        y;

    x = y = s = n = 0;
    fputs ("/pixcol {\n", fp);
    for (j = 0; j < CODELEN; j++) {
        o = pop [i][j].opcode;
        if (opcodes [o].requires <= n) {
            switch (opcodes [o].opt) {
                case END:
                    j = CODELEN;
                    break;
                case NUM:
                    fprintf (fp, "%f\n", pop [i][j].op);
                    n++;
                    break;
                case X:
                    fprintf (
                        fp,
                        "%s%s\n",
                        s ? "Y" : "X",
                        ((x && !s) || (y && s)) ? " neg" : ""
                    );
                    n++;
                    break;
                case Y:
                    fprintf (
                        fp,
                        "%s%s\n",
                        s ? "X" : "Y",
                        ((y && !s) || (x && s)) ? " neg" : ""
                    );
                    n++;
                    break;
                case SWAP:
                    s = !s;
                    break;
                case NEGX:
                    x = !x;
                    break;
                case NEGY:
                    y = !y;
                    break;
                case NOISE:
                    if (! noise)
                        break;
                    /* fall through */
                default:
                    fprintf (fp, "%s\n", opcodes [o].name);
                    n += opcodes [o].changes;
	    }
        }
    }
    fprintf (fp, "} bind def\n");
}

void printcell (int i, FILE *fp)
{
    fprintf (
        fp,
        "1 setgray\n"
        "%i %i 160 160 rectfill\n"
        "%i %i %i %i draw\n",
        (i % 4) * 180 + 20,
        (i / 4) * -180 + 560,
        (i % 4) * 180 + 20,
        (i / 4) * -180 + 560,
        (i % 4) * 180 + 180,
        (i / 4) * -180 + 720
    );
}

void errit (char const *format, ...)
{
    va_list
	list;

    fprintf (stderr, "\nError %s: ", programname);

    va_start (list, format);
    vfprintf (stderr, format, list);

    fprintf (stderr, "\n\n");

    exit (1);
}

void get_programname (char const *argv0)
{
    char
        *p;
    p = strrchr (argv0, '/');
    if (p)
        programname = strdup (p + 1);
    else
        programname = strdup (argv0);
}

void syntax ()
{
    fprintf (
        stderr,
        "\n"
        "Usage: %s [-n]\n"
        "\n"
        "  -n: no noise\n"
        "\n"
        "(c) P. Kleiweg 1998\n"
        "\n",
        programname
    );
    exit (1);
}
